@@ -1,68 +1,83 @@
 """
-This module contains methods and classes for 
-    1. Locating resource classes in specified resources folder
-    2. Loading / Dumping resource instances from / to yaml config files
+This module contains methods and classes for staging resources prior to running experiments.
+
+Resouces are loaded from .yml config files and may be staged remotely.
 """
 
+from __future__ import annotations
+
 import argparse
 import importlib.util
 import inspect
 from pathlib import Path
 import pkgutil
-import time
+from typing import Any
 
 import Pyro5.api as pyro
 import Pyro5.errors
 
+from labctrl.instrument import Instrument
 from labctrl.logger import logger
-from labctrl.resource import Resource, Instrument
+from labctrl.resource import Resource
 from labctrl.settings import Settings
 import labctrl.yamlizer as yml
 
 _PORT = 9090  # port to bind a remote stage on (used to initialize Pyro Daemon)
 _SERVERNAME = "STAGE"
-# remote stage unique resource identifier (URI)
-_STAGE_URI = f"PYRO:{_SERVERNAME}@localhost:{_PORT}"
+_STAGE_URI = f"PYRO:{_SERVERNAME}@localhost:{_PORT}"  # unique resource identifier (URI)
 
 
 class StagingError(Exception):
     """ """
 
 
+class ResourceNotStagedError(Exception):
+    """ """
+
+
 @pyro.expose
 class Stage:
     """ """
 
     def __init__(self, *configpaths: Path, daemon: pyro.Daemon | None = None) -> None:
-        """configpaths: path to YAML file containing Resource classes to be instantiated
-        daemon: if not None, this stage will be a remote stage that serves the Resources in the configpath remotely with Pyro. you can send in multiple configpaths, the resources will be bundled up on the same stage."""
+        """configpaths: path(s) to YAML file containing Resource classes to be instantiated.
+        daemon: if not None, this stage will be a remote stage that serves the Resources in the configpath remotely with Pyro.
+
+        If the stage is initailized with multiple configpaths, all resources will be bundled up on the same stage."""
 
         logger.debug("Initializing a stage...")
 
         # self._config and self._services will be updated by _setup()
         self._config = {}  # dict with key: configpath, value: list of Resources
-
         # if local, services is a dict with key: resource name, value: Resource object
         # if remote, it is a dict with key: resource name, value: remote resource URI
         self._services = {}
 
         self._daemon = daemon
 
+        self._is_local = self._daemon is None
+        mode = "local" if self._is_local else "remote"
         if configpaths:
             self._register()  # locate and register resource classes with yaml and pyro
             self._load(*configpaths)
-            if self._daemon is not None:
+            if not self._is_local:
                 self._serve()
+        else:
+            logger.warning(f"No config(s) supplied, setting up an empty {mode} stage.")
+
+        logger.debug(f"Done initializing a {mode} stage.")
 
     def _register(self) -> None:
         """ """
-        settings = Settings()
-        resourcepath = settings.resourcepath
+        # ensure that the resourcepath is available
+        resourcepath = Settings().resourcepath
+        logger.debug(f"Found labctrl setting {resourcepath = }.")
         if resourcepath is None:
             message = (
                 f"Unable to register resources as no resource path is specified. "
                 f"Please set a resource path through the Settings context manager."
             )
+            logger.error(message)
             raise StagingError(message)
 
         try:
@@ -72,10 +87,19 @@ class Stage:
                 f"Unable to cast {resourcepath = } to a 'Path' object. "
                 f"Please set 'resourcepath' with a string that is a valid path."
             )
+            logger.error(message)
+            raise StagingError(message) from None
+        except RecursionError:
+            message = (
+                f"Unable to setup stage as it is being called by a script within "
+                f"{resourcepath = }. Please call the stage from outside resource path."
+            )
+            logger.error(message)
             raise StagingError(message) from None
         else:
             if not resource_classes:
                 message = "Unable to setup stage as no resource classes were found."
+                logger.error(message)
                 raise StagingError(message)
 
             for resource_class in resource_classes:
@@ -95,19 +119,24 @@ class Stage:
 
             for resource in resources:
                 try:
-                    self._services[resource.name] = resource
+                    resource_name = resource.name
                 except (TypeError, AttributeError):
                     message = (
-                        f"A {resource = } in {configpath = } does not have a 'name'\n"
-                        f"All resources must have a '.name' attribute to be staged"
+                        f"A {resource = } in {configpath = } does not have a 'name'. "
+                        f"All resources must have a '.name' attribute to be staged."
                     )
+                    logger.error(message)
                     raise StagingError(message) from None
+                else:
+                    self._services[resource.name] = resource
+                    logger.info(f"Staged '{resource_name}'.")
 
         if num_resources != len(self._services):
             message = (
-                f"Two or more resources in {configpaths = } share a name\n"
-                f"All resources must have unique names to be staged"
+                f"Two or more resources in {configpaths = } share a name. "
+                f"All resources must have unique names to be staged."
             )
+            logger.error(message)
             raise StagingError(message)
 
     def _serve(self) -> None:
@@ -117,13 +146,22 @@ class Stage:
                 uri = self._daemon.register(resource, objectId=name)
             except (TypeError, AttributeError):
                 message = (
-                    f"Expect {daemon = } of type '{pyro.Daemon}', not {type(daemon)}"
+                    f"Daemon must be of type '{pyro.Daemon}', not {type(self._daemon)}."
                 )
+                logger.error(message)
                 raise StagingError(message) from None
             else:
                 self._services[name] = uri
                 logger.info(f"Served '{resource}' remotely at '{uri}'.")
 
+    def __enter__(self) -> Stage:
+        """ """
+        return self
+
+    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
+        """ """
+        self.teardown()
+
     def save(self) -> None:
         """save current state state to respective yaml configs"""
         logger.debug(f"Saving the current state of staged resources to configs...")
@@ -132,11 +170,48 @@ class Stage:
 
     @property
     def services(self) -> dict[str, str]:
-        """if remote stage, return dict with key = resource name and value = uri
-        if local stage, return dict with key = resource name and value - resource object
+        """if remote stage, store dict with key = resource name and value = uri
+        if local stage, store dict with key = resource name and value = resource object. For remote stages, value will be replaced by Stagehand with a proxy.
         """
         return self._services.copy()
 
+    def replace(self, key: str, value: Any) -> None:
+        """Replace a staged resource with another Resource or Proxy object"""
+        if not isinstance(value, (Resource, pyro.Proxy)):
+            message = (
+                f"Only Resources and Proxies can be staged, not {value} "
+                f"of {type(value)}."
+            )
+            logger.error(message)
+            raise StagingError(message)
+
+        try:
+            self._services[key] = value
+        except KeyError:
+            message = f"No resource named '{key}' has been staged."
+            logger.error(message)
+            raise ResourceNotStagedError(message) from None
+        else:
+            logger.debug(f"Replaced resource named '{key}' with {value}.")
+
+    def get(self, *resources: str) -> Resource | list[Resource]:
+        """ """
+        if not resources:
+            return []
+        try:
+            if len(resources) == 1:
+                (resource,) = resources
+                return self._services[resource]
+            else:
+                return [self._services[resource] for resource in resources]
+        except KeyError:
+            message = (
+                f"One or more names in {resources} do not exist on this stage. "
+                f"Please check if you have staged all resources you are trying to get."
+            )
+            logger.error(message)
+            raise ResourceNotStagedError(message) from None
+
     def teardown(self) -> None:
         """disconnect instruments (if any) and shutdown daemon request loop if remote"""
         self.save()
@@ -165,30 +240,25 @@ class Stagehand:
         logger.debug("Initializing a stagehand...")
 
         self._stage = Stage(*configpaths)
-        # set resource names as stage attributes for easy access
-        for name, resource in self._stage.services.items():
-            setattr(self._stage, name, resource)
-            logger.info(f"Set stage attribute '{name}'.")
 
         # attributes '_remote_stage' and '_proxies' are updated if remote stage exists
         self._proxies: list[pyro.Proxy] = []
         self._remote_stage: pyro.Proxy = pyro.Proxy(_STAGE_URI)
+
         try:
             for name, uri in self._remote_stage.services.items():
                 proxy = pyro.Proxy(uri)
                 self._proxies.append(proxy)
-                setattr(self._stage, name, proxy)
-                logger.info(
-                    f"Set local stage attribute '{name}' after finding a remote "
-                    f"resource served at {uri}."
-                )
+                self._stage.replace(name, proxy)  # replace resource with proxy
+                logger.info(f"Found and staged a remote resource served at {uri}.")
         except Pyro5.errors.CommunicationError:
-            logger.warning(
+            message = (
                 f"Did not find a remote stage at {_STAGE_URI}. "
-                f"Ignore this warning if you are only using local resources. "
-                f"Please setup a remote stage if you intend to use remote resources."
+                f"Please setup a remote stage if you want to use remote resources. "
+                f"Use a Stage context manager if you want to only use local resources."
             )
-            self._remote_stage = None
+            logger.error(message)
+            raise StagingError(message) from None
 
     @property
     def stage(self) -> Stage:
@@ -229,21 +299,48 @@ def locate(source: Path) -> set[Resource]:
     return resources
 
 
-def validate(configpaths: list[Path]) -> None:
+def setup(*configpaths: Path) -> None:
     """ """
-    if not configpaths:
-        raise StagingError(
-            f"Failed to setup stage as no configpaths were provided. "
-            f"Please provide at least one path to a yml config file and try again."
-        )
+    logger.info("Setting up a remote stage...")
     for path in configpaths:
-        if path.suffix not in (".yml", ".yaml"):
-            raise StagingError(
-                f"Unrecognized configpath '{path.name}'. "
-                f"Valid configs are YAML files with a '.yml' or '.yaml' extension."
-            )
+        try:
+            if path.suffix.lower() not in (".yml", ".yaml"):
+                message = (
+                    f"Unrecognized configpath '{path.name}'. "
+                    f"Valid configs are YAML files with a '.yml' or '.yaml' extension."
+                )
+                logger.error(message)
+                raise StagingError(message)
+            logger.debug(f"Validated all {configpaths = }")
+        except (TypeError, AttributeError):
+            message = f"Expect path of '{Path}', not '{path}' of '{type(path)}'"
+            logger.error(message)
+            raise StagingError(message) from None
 
-    logger.debug(f"Validated all {configpaths = }")
+    # create pyro Daemon and register a remote stage
+    daemon = pyro.Daemon(port=_PORT)
+    stage = Stage(*configpaths, daemon=daemon)
+    stage_uri = daemon.register(stage, objectId=_SERVERNAME)
+    logger.info(f"Served a remote stage at '{stage_uri}'.")
+
+    # start listening for requests
+    with daemon:
+        logger.info("Remote stage daemon is now listening for requests...")
+        daemon.requestLoop()
+        logger.info("Exited remote stage daemon request loop.")
+
+
+def teardown() -> None:
+    """ """
+    try:
+        with pyro.Proxy(_STAGE_URI) as remote_stage:
+            remote_stage.teardown()
+    except Pyro5.errors.CommunicationError:
+        message = f"No remote stage to teardown at {_STAGE_URI}. "
+        logger.error(message)
+        raise StagingError(message) from None
+    else:
+        logger.info(f"Tore down a remote stage at '{_STAGE_URI}'.")
 
 
 if __name__ == "__main__":
@@ -265,26 +362,17 @@ if __name__ == "__main__":
 
     # command line argument handling
     if args.run:  # setup remote stage with resources from user supplied configpath(s)
-        logger.info("Setting up a remote stage...")
-
         # extract configpaths from args
         configpaths = [Path(configpath) for configpath in args.configpaths]
-        validate(configpaths)
+        if not configpaths:
+            message = (
+                f"Failed to setup stage as no configpaths were provided. "
+                f"Please provide at least one path to a yml config file and try again."
+            )
+            logger.error(message)
+            raise StagingError(message)
 
-        # create pyro Daemon and register a remote stage
-        daemon = pyro.Daemon(port=_PORT)
-        stage = Stage(*configpaths, daemon=daemon)
-        stage_uri = daemon.register(stage, objectId=_SERVERNAME)
-        logger.info(f"Served a remote stage at '{stage_uri}'.")
-
-        # start listening for requests
-        with daemon:
-            logger.info("Remote stage daemon is now listening for requests...")
-            daemon.requestLoop()
-            logger.info("Exited remote stage daemon request loop.")
+        setup(*configpaths)
 
     else:  # teardown remote stage
-        logger.info("Tearing down remote stage at '{_STAGE_URI}', please wait ~2s...")
-        with pyro.Proxy(_STAGE_URI) as remote_stage:
-            remote_stage.teardown()
-        time.sleep(2)
+        teardown()
